// This class instatiates the E-Z Reader model of eye-movement control during 
// reading (ver. 10.2; see Reichle, 2011, Oxford Handbook of Eye Movements).  

package reader;

import java.io.PrintStream;
import java.util.*;

public class EZReader {
   
    public void run(PrintStream diskWriter, ArrayList<Sentence> text, Utility utility) {
        
        // Initialize classes:
        Display display = new Display();
        Random r = new Random();
        Stats stats = new Stats();
        
        // Note: This array records the number or regressions.  
        int regressionN[] = new int[Reader.NSentences];
        
        for (int subject = 0; subject < Reader.NSubjects; subject++) {
           
            for (int S = 0; S < Reader.NSentences; S++) {

                // Initialize fixation:
                int fixationN = 0;
                ArrayList<Fixation> trace = new ArrayList<>();
                Fixation f = new Fixation();
                f.number = fixationN;
                f.dur = 0;
                f.pos = text.get(S).get(0).iv.OVP;
                f.word = 0;
                
                // Start L1:
                ArrayList<Process> active = new ArrayList<>();
                Process p = new Process();
                int N = 0;
                p.initializeL1(text, S, N, true);
                double processingRate = p.calcProcessingRate(text, S, N, f.pos);
                p.duration *= processingRate;
                active.add(p);
                
                // Read sentence:
                boolean sentenceDone = false;
                while (!sentenceDone) {
                    
                    // *** Display model state:
                    if (Reader.displayModelStates == true) display.state(diskWriter, S, N, f.pos, processingRate, active);
                    
                    // Identify processing w/ shortest duration:
                    double minProcessDuration = active.get(0).duration;
                    int minProcessID = 0;
                    for (int i = active.size() - 1; i >= 0; i--) if (active.get(i).duration < minProcessDuration) {
                        minProcessDuration = active.get(i).duration;
                        minProcessID = i;
                    }
                    
                    // Store attributes of shortest process:
                    Process sp = new Process();
                    sp.duration = active.get(minProcessID).duration;
                    sp.durationCopy = active.get(minProcessID).durationCopy;
                    sp.name = active.get(minProcessID).name;
                    sp.saccadeLength = active.get(minProcessID).saccadeLength;
                    sp.word = active.get(minProcessID).word;
                    
                    // Remove shortest process from list of active processes:
                    active.remove(minProcessID);
                    
                    // Decrement all remaining process durations:
                    for (int i = 0; i < active.size(); i++) active.get(i).duration -= sp.duration;
                            
                    // Increase fixation duration:
                    if (!sp.name.equals("S")) f.dur += sp.duration;
                    
                    //**********************************************************
                    // DETERMINE NEXT MODEL STATE
                    //**********************************************************
                    
                    // PRE-ATTENTIVE VISUAL PROCESSING:
                    if (sp.name.equals("V")) {
                        
                        // Adjust L1 rate:
                        for (int i = 0; i < active.size(); i++) if (active.get(i).name.equals("L1")) {
                            
                            active.get(i).duration /= processingRate;                            
                            processingRate = p.calcProcessingRate(text, S, N, f.pos);                       
                            active.get(i).duration *= processingRate;                           
                        }
                    }
                    
                    // FAMILIARITY CHECK:
                    else if (sp.name.equals("L1")) {
                        
                        // Deterine if integration is on-going:
                        boolean ongoingI = false;
                        for (int i = 0; i < active.size(); i++) if (active.get(i).name.equals("I") && active.get(i).word == (N - 1)) ongoingI = true;
                        
                        // Start L2:
                        p = new Process();
                        p.initializeL2(text, S, N, ongoingI);
                        active.add(p);
                        
                        // Cancel any pending M1 & start M1:
                        if (N < text.get(S).numberWords - 1) {
                            p = new Process();
                            p.initializeM1(active, f.pos, text.get(S).get(N + 1).iv.OVP, N + 1);
                            active.add(p);
                        }
                    }
                    
                    // LEXICAL ACCESS:
                    else if (sp.name.equals("L2")) {
                        
                        // Determine if integation of the previous word failed:
                        boolean IFailure = false;
                        for (int i = 0; i < active.size(); i++) if (active.get(i).name.equals("I")) {
                            IFailure = true;
                            N = active.get(i).word; // i.e., location of integration failure
                            active.remove(i);
                        }
                        
                        // Integration failed:
                        if (IFailure == true) {
                            //System.out.printf(" Slow integration failure\n");
                            
                            // Start A:
                            p = new Process();
                            p.initializeA();
                            active.add(p);
                            
                            // Cancel any pending M1 & start M1:
                            p = new Process();
                            p.initializeM1(active, f.pos, text.get(S).get(N).iv.OVP, N);
                            active.add(p);

                        }
                        
                        // Integration was successful:
                        else { 
                            
                            // Start I:
                            p = new Process();
                            p.initializeI(N);
                            active.add(p);
                        
                            // Start A:
                            if (N == text.get(S).numberWords - 1) sentenceDone = true;
                            else {
                                p = new Process();
                                p.initializeA();
                                active.add(p);
                            } 
                        }
                    }
                    
                    // POST-LEXICAL INTEGRATION:
                    else if (sp.name.equals("I")) {  
                        
                        // Rapid integration failure:
                        double PrIFailure = r.nextDouble();
                        if (N > 0 && PrIFailure < Reader.pF) {
                            //System.out.printf(" Rapid integration failure\n");
                            
                            for (int i = 0; i < active.size(); i++) if (active.get(i).name.equals("A")) active.remove(i);
                            for (int i = 0; i < active.size(); i++) if (active.get(i).name.equals("L1")) active.remove(i);
                            for (int i = 0; i < active.size(); i++) if (active.get(i).name.equals("L2")) active.remove(i);
                            
                            // Start A:
                            N = sp.word; // i.e., shift attention to source of integration difficulty
                            p = new Process();
                            p.initializeA();
                            active.add(p);
                            
                            // Cancel any pending M1 & start M1:
                            p = new Process();
                            p.initializeM1(active, f.pos, text.get(S).get(N).iv.OVP, N);
                            active.add(p);
                        }
                    }
                    
                    // ATTENTION SHIFT:
                    else if (sp.name.equals("A")) {
                        
                        // Shift attention to next word:
                        N++;
                        
                        // Deterine if integration is on-going:
                        boolean ongoingI = false;
                        for (int i = 0; i < active.size(); i++) if (active.get(i).name.equals("I") && active.get(i).word == (N - 1)) ongoingI = true;
                        
                        // Start L1:
                        p = new Process();
                        p.initializeL1(text, S, N, ongoingI);
                        processingRate = p.calcProcessingRate(text, S, N, f.pos);
                        p.duration *= processingRate;
                        active.add(p);
                    }
                    
                    // LABILE SACCADIC PROGRAMMING:
                    else if (sp.name.equals("M1")) {
                        
                        // Start M2:
                        p = new Process();
                        p.initializeM2(sp.saccadeLength, f.dur, sp.word);
                        active.add(p);
                    }
                    
                    // NON-LABILE SACCADIC PROGRAMMING:
                    else if (sp.name.equals("M2")) {
                        
                        // Start S:
                        p = new Process();
                        p.initializeS(sp.saccadeLength, sp.word);
                        active.add(p);
                    }
                    
                    // SACCADE EXECUTION:
                    else if (sp.name.equals("S")) {

                        // Terminate previous fixation:
                        trace.add(f);
                        
                        // Start new fixation:
                        fixationN++;
                        f = new Fixation();
                        f.number = fixationN;
                        f.dur = 0;
                        f.pos = trace.get(fixationN - 1).pos + sp.saccadeLength;
                        if (f.pos < 0) f.pos = 0;
                        double lastChar = text.get(S).get(text.get(S).numberWords - 1).iv.charPosN;
                        if (f.pos > lastChar) f.pos = lastChar; 
                        for (int i = 0; i < text.get(S).numberWords; i++) if (f.pos >= text.get(S).get(i).iv.charPos0 && f.pos < text.get(S).get(i).iv.charPosN) f.word = i;
                        //System.out.printf(" %5.4f %2d\n", f.pos, f.word);
                       
                        // Start V:
                        p = new Process();
                        p.initializeV(N);
                        active.add(p);
                        
                        // Start M1 (automatic refixation):
                        double PrRefixate = r.nextDouble();
                        double saccadeError = Math.abs(text.get(S).get(sp.word).iv.OVP - f.pos);
                        if (PrRefixate < (Reader.Lambda * saccadeError)) {
                            p = new Process();
                            p.initializeM1(active, f.pos, text.get(S).get(sp.word).iv.OVP, sp.word);
                            active.add(p);
                        }
                    }
                    
                } // End of read-sentence loop
               
                active.clear();
                
                if (Reader.displayTrace == true) display.fixations(diskWriter, text, S, trace);

                if (Reader.includeRegressions == false) {
                    boolean regression = false;
                    for (int i = 1; i < trace.size(); i++) if (trace.get(i - 1).word > trace.get(i).word) regression = true;
                    if (regression == false) stats.calcWordBasedMeasures(text, S, trace);
                    else regressionN[S]++;
                }
                else stats.calcWordBasedMeasures(text, S, trace);
                
                if (Reader.displayWordBasedMeasures == true) display.displayWordBasedMeasures(diskWriter, text, S);
                
                utility.calcMeans(trace, utility);
                
                trace.clear();
                
            } // End of sentences loop

        } // End of subject loop

        stats.calcWordBasedMeans(text, regressionN);
        
        if (Reader.displayWordBasedMeans == true) {
            display.wordBasedMeans(diskWriter, text);
            display.fixDists(diskWriter, text);
            display.refixProbs(diskWriter, text);
            display.IOVPs(diskWriter, text);
        }
        
    } // End of run model loop 
}
