// This class contains a number of methods that are used to convert fixations 
// traces into a variety of word-based measures, including both individual words 
// and, e.g., collapsed across words of the same length (e.g., first-fixation 
// landing-site distributions).

package reader;

import java.util.*;

public class Stats {
    
    //**************************************************************************
     
    // Update word-based eye-movement measures for each word in sentence:
    public void calcWordBasedMeasures(ArrayList<Sentence> text, int S, ArrayList<Fixation> trace) {
        
        for (int i = 1; i < text.get(S).numberWords - 1; i++) {
            
            // Set up word class to hold DV values:
            Word w = new Word();
      
            // Convert trace fixations into DVs for each word:
            for (int j = 0; j < trace.size(); j++) {
                if (trace.get(j).word == i) {
                    boolean regression = false;
                    for (int k = 0; k <= j; k++) if (trace.get(k).word > i) regression = true;
                    if (regression == true) {
                        w.dv.NFix++;
                        w.dv.TT += trace.get(j).dur;
                    }
                    else {
                        w.dv.NFix++;
                        w.dv.NFPFix++;
                        if (w.dv.FFD == 0) {
                            w.dv.FFD = trace.get(j).dur;
                            w.dv.fixPos = (int)(trace.get(j).pos - text.get(S).get(trace.get(j).word).iv.charPos0); 
                            w.dv.FixDist[w.dv.fixPos]++; // fixation distribution
                            if (j < (trace.size() - 1) && i == trace.get(j + 1).word) w.dv.RefixProb[w.dv.fixPos]++;
                            w.dv.GD = trace.get(j).dur;
                        }
                        else if (j == 0 || (j > 0 && trace.get(j - 1).word == i)) w.dv.GD += trace.get(j).dur;
                        w.dv.TT += trace.get(j).dur;
                    }
                }
            }
            
            // Calculate DVs that cannot be directly derived from trace:
            if (w.dv.FFD == 0) w.dv.PrS++;
            else {
                w.dv.PrF++;
                if (w.dv.FFD == w.dv.GD) {
                    w.dv.SFD = w.dv.FFD;
                    w.dv.Pr1++;
                    // Calculate IOVP:
                    w.dv.IOVP[w.dv.fixPos] = w.dv.SFD;
                    w.dv.fixPosN[w.dv.fixPos]++;
                }
                else w.dv.Pr2++;
            }

            // Add values to running totals used for means across subjects:
            text.get(S).get(i).dv.SFD += w.dv.SFD;
            text.get(S).get(i).dv.FFD += w.dv.FFD;
            text.get(S).get(i).dv.GD += w.dv.GD;
            text.get(S).get(i).dv.TT += w.dv.TT;
            text.get(S).get(i).dv.Pr1 += w.dv.Pr1;
            text.get(S).get(i).dv.Pr2 += w.dv.Pr2;
            text.get(S).get(i).dv.PrS += w.dv.PrS;
            text.get(S).get(i).dv.PrF += w.dv.PrF;
            for (int j = 0; j < Reader.maxWordLength; j++) {
                text.get(S).get(i).dv.FixDist[j] += w.dv.FixDist[j];
                text.get(S).get(i).dv.RefixProb[j] += w.dv.RefixProb[j];
                text.get(S).get(i).dv.IOVP[j] += w.dv.IOVP[j];
                text.get(S).get(i).dv.fixPosN[j] += w.dv.fixPosN[j];
            }

        } 
    }
    
    //**************************************************************************
     
    // Calculate word-based means for each word in text:
    public void calcWordBasedMeans(ArrayList<Sentence> text, int[] regressionN) {
        
        // Fixation duration measures:
        for (int i = 0; i < Reader.NSentences; i++) {
            for (int j = 1; j < text.get(i).numberWords - 1; j++) {
            
                if (text.get(i).get(j).dv.Pr1 > 0) text.get(i).get(j).dv.SFD /= text.get(i).get(j).dv.Pr1;
                if ((text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2) > 0) {
                    text.get(i).get(j).dv.FFD /= (text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2);
                    text.get(i).get(j).dv.GD /= (text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2);
                    text.get(i).get(j).dv.TT /= (text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2);
                    
                    // Distributions:
                    for (int k = 0; k < Reader.maxWordLength; k++) {
                        if (text.get(i).get(j).dv.FixDist[k] > 0) text.get(i).get(j).dv.RefixProb[k] /= text.get(i).get(j).dv.FixDist[k];
                        text.get(i).get(j).dv.FixDist[k] /= (text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2);                        
                        if (text.get(i).get(j).dv.fixPosN[k] > 0) text.get(i).get(j).dv.IOVP[k] /= text.get(i).get(j).dv.fixPosN[k];
                    }
                }
             
                // Fixation probability measures:
                text.get(i).get(j).dv.Pr1 /= (Reader.NSubjects - regressionN[i]);
                text.get(i).get(j).dv.Pr2 /= (Reader.NSubjects - regressionN[i]);
                text.get(i).get(j).dv.PrS /= (Reader.NSubjects - regressionN[i]);
                text.get(i).get(j).dv.PrF /= (Reader.NSubjects - regressionN[i]);
            }
        }
    }
    
    //**************************************************************************
     
    // Calculate word-based means for 5 frequency classes of words:
    public void calcClassMeans(ArrayList<Sentence> text, ArrayList<Word> classMeans) {
        
        // Add up DV means for each frequency class:
        for (int i = 0; i < Reader.NSentences; i++) {
            for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                int freqClass = text.get(i).get(j).iv.freqClass;
                classMeans.get(freqClass).dv.SFD += text.get(i).get(j).dv.SFD;
                classMeans.get(freqClass).dv.FFD += text.get(i).get(j).dv.FFD;
                classMeans.get(freqClass).dv.GD += text.get(i).get(j).dv.GD;
                classMeans.get(freqClass).dv.Pr1 += text.get(i).get(j).dv.Pr1;
                classMeans.get(freqClass).dv.Pr2 += text.get(i).get(j).dv.Pr2;
                classMeans.get(freqClass).dv.PrS += text.get(i).get(j).dv.PrS;
                classMeans.get(freqClass).dv.N++;
            }
        }
        
        // Calculate frequency-class means:
        for (int i = 0; i < 5; i++) {
            classMeans.get(i).dv.SFD /= classMeans.get(i).dv.N;
            classMeans.get(i).dv.FFD /= classMeans.get(i).dv.N;
            classMeans.get(i).dv.GD /= classMeans.get(i).dv.N;
            classMeans.get(i).dv.Pr1 /= classMeans.get(i).dv.N;
            classMeans.get(i).dv.Pr2 /= classMeans.get(i).dv.N;
            classMeans.get(i).dv.PrS /= classMeans.get(i).dv.N;
        }
    }
    
    //**************************************************************************
     
    // Calculate means first-fixation landing-site distributions:
    public void calcFixDists(ArrayList<Sentence> text, ArrayList<Word> fixDists) {
        
        // Calculate distributions:
        for (int wordLength = 0; wordLength < Reader.maxWordLength; wordLength++) {
            
            for (int i = 0; i < Reader.NSentences; i++) {
                for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                    if (text.get(i).get(j).iv.length == wordLength) {
                        if (text.get(i).get(j).dv.Pr1 != 0) {
                            for (int k = 0; k < Reader.maxWordLength; k++) fixDists.get(wordLength).dv.FixDist[k] += text.get(i).get(j).dv.FixDist[k];
                            fixDists.get(wordLength).dv.N++;
                        }
                    }
                }
            }
        }
        for (int wordLength = 0; wordLength < Reader.maxWordLength; wordLength++) {
            for (int i = 0; i < Reader.maxWordLength; i++) if (fixDists.get(wordLength).dv.N > 0) fixDists.get(wordLength).dv.FixDist[i] /= fixDists.get(wordLength).dv.N;
        }
    }
    
    //**************************************************************************
     
    // Calculate means refixation landing-site distributions:
    public void calcRefixProbs(ArrayList<Sentence> text, ArrayList<Word> refixProbs) {
        
        // Caclulate distributions:
        for (int wordLength = 0; wordLength < Reader.maxWordLength; wordLength++) {
            for (int i = 0; i < Reader.NSentences; i++) {
                for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                    if (text.get(i).get(j).iv.length == wordLength) {
                        for (int k = 0; k < Reader.maxWordLength; k++) refixProbs.get(wordLength).dv.RefixProb[k] += text.get(i).get(j).dv.RefixProb[k];
                        refixProbs.get(wordLength).dv.N++;
                    }
                }
            }
        }
        for (int wordLength = 0; wordLength < Reader.maxWordLength; wordLength++) {
            for (int i = 0; i < Reader.maxWordLength; i++) if (refixProbs.get(wordLength).dv.N > 0) refixProbs.get(wordLength).dv.RefixProb[i] /= refixProbs.get(wordLength).dv.N;
        }
    }

    //**************************************************************************
     
    public void calcIOVPs(ArrayList<Sentence> text, ArrayList<Word> IOVPs) {
        
        // Calculate distributions:
        for (int wordLength = 0; wordLength < Reader.maxWordLength; wordLength++) {
            
            for (int i = 0; i < Reader.NSentences; i++) {
                for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                    if (text.get(i).get(j).iv.length == wordLength) {
                        for (int k = 0; k < Reader.maxWordLength; k++) {
                            if (text.get(i).get(j).dv.IOVP[k] > 0) {
                                IOVPs.get(wordLength).dv.IOVP[k] += text.get(i).get(j).dv.IOVP[k];
                                IOVPs.get(wordLength).dv.fixPosN[k]++;
                            }
                        }
                    }
                }
            }
        }
        for (int wordLength = 0; wordLength < Reader.maxWordLength; wordLength++) {
            for (int i = 0; i < Reader.maxWordLength; i++) if (IOVPs.get(wordLength).dv.fixPosN[i] > 0) IOVPs.get(wordLength).dv.IOVP[i] /= IOVPs.get(wordLength).dv.fixPosN[i];
        }
    }
    
    //**************************************************************************
     
    // Calcuate RMSD for Schilling et al. (1998) word-frequency classes:
    double calcRMSD(ArrayList<Word> classMeans) {
        double result;
        
        // Initialize observed means and standard deviations (based on Schilling et al., 1998):
        double[] obsMeanFFD = new double[] { 248, 234, 228, 223, 208 }; 
        double[] obsSDFFD = new double[] { 26.34, 31.65, 31.31, 42.21, 64.75 }; 
        double[] obsMeanGD = new double[] { 293, 272, 256, 234, 214 }; 
        double[] obsSDGD = new double[] { 38.76, 44.81, 40.29, 56.61, 71.11 }; 
        double[] obsMeanPr1 = new double[] { 0.68, 0.70, 0.68, 0.44, 0.32 }; 
        double[] obsSDPr1 = new double[] { 0.4665, 0.4583, 0.4665, 0.4964, 0.4665 }; 
        double[] obsMeanPrS = new double[] { 0.10, 0.13, 0.22, 0.55, 0.67 };  
        double[] obsSDPrS = new double[] { 0.3000, 0.3363, 0.4142, 0.4975, 0.4702 }; 
        
        // Calculate RMSD:
        double RMSD = 0;
        for (int i = 0; i < 5; i++) {
            RMSD += Math.pow(((obsMeanFFD[i] - classMeans.get(i).dv.FFD) / obsSDFFD[i]), 2);
            RMSD += Math.pow(((obsMeanGD[i] - classMeans.get(i).dv.GD) / obsSDGD[i]), 2);
            RMSD += Math.pow(((obsMeanPr1[i] - classMeans.get(i).dv.Pr1) / obsSDPr1[i]), 2);
            RMSD += Math.pow(((obsMeanPrS[i] - classMeans.get(i).dv.PrS) / obsSDPrS[i]), 2);
        }
        RMSD /= 20.0; // Note: 20 pairs of means
        result = Math.sqrt(RMSD);
        return result;
    }
}
